package es.caib.signatura.client.swing;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.nio.charset.Charset;
import java.security.InvalidAlgorithmParameterException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.cert.CertStoreException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import javax.mail.MessagingException;

import org.bouncycastle.cms.CMSException;
import org.bouncycastle.x509.NoSuchStoreException;

import es.caib.signatura.api.SMIMEParser;
import es.caib.signatura.api.Signature;
import es.caib.signatura.api.SignatureCertNotFoundException;
import es.caib.signatura.api.SignatureDataException;
import es.caib.signatura.api.SignatureException;
import es.caib.signatura.api.SignaturePrivKeyException;
import es.caib.signatura.api.SignatureVerifyException;
import es.caib.signatura.impl.CMSSignatureRawv2;
import es.caib.signatura.impl.CMSSignaturev2;
import es.caib.signatura.impl.ClassLoaderFactory;
import es.caib.signatura.impl.MIMEInputStream;


/**
 * Esta clase implementa un circuito de prueba completo para un certificado.
 */

public class CircuitoDePrueba {
	private class CircuitException extends Throwable{
		public int code=1;
		public String descripcion=null;
		
		public CircuitException(int code,Throwable cause) {
			super(cause);
			this.code = code;
			this.descripcion=descripcion;
		}
	}
	
	private String contentType="application/signaturaTestNR";
	private es.caib.signatura.api.Signer signer = null;  
	private	String[] systemCertList=null;
	private String version="";
	
	private	HashMap circuitCertList=new HashMap();
	private StringBuffer allCertNames=new StringBuffer("");
	private Vector sigVec=new Vector(); 
	private String inputPath=".";
	private String outputPath=".";
	private int fases;
	private Vector errores=new Vector();
	private static CircuitoDePrueba circuit;
	
	
	public static void main(String[] args) {
		
		circuit=new CircuitoDePrueba(args);		

		int status=0;
		
		circuit.run();

		int i=0;

		for(i=0;i<circuit.errores.size();i++){
			try{
				System.out.println("Error "+i+" : "+((Exception)circuit.errores.get(i)).getMessage());
				((Exception)circuit.errores.get(i)).printStackTrace();
			}catch(Exception ex){}
		}
		if(circuit.errores.size()!=0){
			System.exit(1);
		}else{
			System.exit(0);
		}
		
		
		
	}
	


	private void run()   {
		//preprocessing global
		File pdf_firma_multiple=new File(outputPath+"/pdf_firmado_multiple_"+version+".pdf");
		
		if(pdf_firma_multiple.exists())pdf_firma_multiple.delete();
		
		
		//Por cada certificado que ha introducido el usuario
		//ejecutamos el circuito de prueba
	    Iterator it=circuitCertList.keySet().iterator();
	    while(it.hasNext()){
    		String circuitItem=(String)it.next();
	    	allCertNames.append("_"+circuitItem.replaceAll(":", "")); //preparamos el nombre para las firmas en paralelo
	    	int i;
	    	
    		
			contentType="application/signaturaTest";
			run(circuitItem,(String)circuitCertList.get(circuitItem),"CR_SA");
			
			contentType="application/signaturaTestNR";
			run(circuitItem,(String)circuitCertList.get(circuitItem),"CNR_SA");
			
			contentType="application/signaturaTestTimeStamping";
			run(circuitItem,(String)circuitCertList.get(circuitItem),"CR_SAST");
			
			contentType="application/x-caib-rebutregistre";
			run(circuitItem,(String)circuitCertList.get(circuitItem),"CNR_SAST");
			
			contentType="application/signaturaTest";
			signPDFMultiple(circuitItem,(String)circuitCertList.get(circuitItem),"CNR_SA");
			
			
    	}
	    
	    //postprocessing global
	    //Generamos el SMIME con firmas paralelas correspondiente al documento
	    String inputFileName="documento_de_prueba.txt";
	    String baseOutputFileName=outputPath+"/"+version+"_"+
	  		System.getProperty("os.name")+"_"+
	  		System.getProperty("java.specification.version")+allCertNames.toString();
	    

	    
	    InputStream smimein=null;
		OutputStream smimeout=null;
		try {
			smimein = new FileInputStream(inputPath+"/"+inputFileName);
			smimeout = new FileOutputStream(baseOutputFileName+inputFileName+".smime");
			signer.generateSMIMEParalell(smimein, (Signature [])sigVec.toArray(new Signature[sigVec.size()]), smimeout);
			System.err.println("SMIME PARALELO generado");
		} catch (Exception e) {		
			e.printStackTrace();
			errores.add(new CircuitException(310,e));
		}finally{
			try {
				if(smimein!=null)smimein.close();
			} catch (IOException e) {	
			}
			try {
				if(smimeout!=null)smimeout.close();
			} catch (IOException e) {
			}
		
		}
		
	    sigVec=new Vector();
		
		it=circuitCertList.keySet().iterator();
	    while(it.hasNext()){
    		String circuitItem=(String)it.next();
	    	allCertNames.append("_"+circuitItem.replaceAll(":", "")); //preparamos el nombre para las firmas en paralelo
	    	int i;
	
			//PROVES DE SIGNATURA RAW
			contentType="application/caib-raw-sign-recognized";
			runRAW(circuitItem,(String)circuitCertList.get(circuitItem),"CR_SA_RAW");
			
			contentType="application/caib-raw-sign";
			runRAW(circuitItem,(String)circuitCertList.get(circuitItem),"CNR_SA_RAW");
			
			contentType="application/caib-raw-sign-recognized-timestamp";
			runRAW(circuitItem,(String)circuitCertList.get(circuitItem),"CR_SAST_RAW");
			
			contentType="application/caib-raw-sign-timestamp";
			runRAW(circuitItem,(String)circuitCertList.get(circuitItem),"CNR_SAST_RAW");

			
			
			 //postprocessing global
		    //Generamos el SMIME con firmas paralelas correspondiente al documento
		    inputFileName="documento_de_prueba.txt";
		    baseOutputFileName=outputPath+"/RAW_"+version+"_"+
		  	System.getProperty("os.name")+"_"+
		  	System.getProperty("java.specification.version")+allCertNames.toString();
		    
		    smimein=null;
			smimeout=null;
			try {
				smimein = new FileInputStream(inputPath+"/"+inputFileName);
				smimeout = new FileOutputStream(baseOutputFileName+inputFileName+".smime");
				signer.generateSMIMEParalell(smimein, (Signature [])sigVec.toArray(new Signature[sigVec.size()]), smimeout);
				System.err.println("SMIME PARALELO generado");
			} catch (Exception e) {		
				e.printStackTrace();
				errores.add(new CircuitException(310,e));
			}finally{
				try {
					if(smimein!=null)smimein.close();
				} catch (IOException e) {	
				}
				try {
					if(smimeout!=null)smimeout.close();
				} catch (IOException e) {
				}
			
			}
	    }
	}

	
	private void signPDFMultiple(String certificateName,String password,String additionalAttrib) {
		try{
		      //Firmamos un PDF
			String inputFileName="documento_de_prueba.pdf";
			String baseOutputFileName=outputPath+"/pdf_firmado_multiple_"+version+".pdf";

	  
		      //firmamos el pdf
			InputStream pdfin;
			if(!new File(baseOutputFileName).exists())
				pdfin=new FileInputStream(inputPath+"/"+inputFileName);
			else
				pdfin=new FileInputStream(baseOutputFileName);
			
			OutputStream pdfout=new FileOutputStream(baseOutputFileName+".tmp"); //new File(baseOutputFileName+inputFileName).canRead()
		    String url="www.caib.es";
		    int position=signer.PDF_SIGN_POSITION_NONE;
		    signer.signPDF(pdfin,pdfout, certificateName, password, contentType, url, position,true);
		    System.err.println("PDF MULTIPLE firmado");
		      
		    pdfin.close();
		    pdfout.close();
		    
		    copy(baseOutputFileName+".tmp",baseOutputFileName);
		    
		    
		}catch(Exception e){
			errores.add(e);
			e.printStackTrace();
		}
		
	}



	private void copy(String inPath, String outPath) throws Exception{
		InputStream in= new BufferedInputStream(new FileInputStream(inPath));
		OutputStream out=new BufferedOutputStream(new FileOutputStream(outPath));
		
		byte  readed []=new byte[10240];
		int read=0;
		while((read=in.read(readed))!=-1)
			out.write(readed,0,read);
		
		in.close();
		out.close();
	}



	private void run(String certificateName,String password,String additionalAttrib) {
		 //Ejecutamos el circuito para un certificado en concreto
		System.err.println("********** RUN ***************************");
		System.err.println("content-type: "+contentType);	
		System.err.println("certificado para firmar: "+certificateName);
		System.err.println("atributos: "+additionalAttrib);		
		
		try {
			check_certs();
		} catch (CircuitException e1) {
			errores.add(e1);
			e1.printStackTrace();
		}
		
		String baseOutputFileName=outputPath+"/"+version+"_"+
  		System.getProperty("os.name")+"_"+
  		System.getProperty("java.specification.version")+"_"+additionalAttrib+"_"+certificateName.replaceAll(":", "").replaceAll("\\\\", "_")+"_";
		
		String inputFileName="documento_de_prueba.txt";
		
		try {
		

			 //Firmamos el documento de prueba
			  
			  
			  Signature signatureData=signer.sign(inputPath+"/"+inputFileName, certificateName, password, contentType);
			  sigVec.add(signatureData);
			  
			  //Generamos el fichero de firma
		      OutputStream oout = new FileOutputStream(baseOutputFileName+inputFileName+".pkcs7");
		      oout.write(signatureData.getPkcs7());
		      oout.close();
		      System.err.println("Firma generada");
		      
		      //Validamos el documento con el fichero de firma

		      if(signatureData.verify(new FileInputStream(inputPath+"/"+inputFileName))){
		    	  System.err.println("Firma verificada"); 
		      }else{
		    	  System.err.println("Error en la validacion");
		      }

		      if(signatureData.verify()){
		    	  System.err.println("Certificado válido"); 
		      }else{
		    	  System.err.println("Certificado no valido");
		      }
		      

		      //Generamos el SMIME correspondiente al documento
		      String smimeFileName=baseOutputFileName+inputFileName+".p7s.slc";
		      InputStream smimein=new FileInputStream(inputPath+"/"+inputFileName);
		      OutputStream smimeout=new FileOutputStream(smimeFileName);
		      signer.generateSMIME(smimein, signatureData, smimeout);
		      smimein.close();
		      smimeout.close();
		      System.err.println("SMIME generado");
		      
		      //verificamos el smime
		      verifySmime(smimeFileName);
		}catch(Exception e){
			errores.add(e);
			e.printStackTrace();
		}
		
		try {
			
		      //Firmamos un PDF
			  inputFileName = "documento_de_prueba.pdf";
	  
		      //firmamos el pdf
		      InputStream pdfin = new FileInputStream(inputPath+"/"+inputFileName);
		      OutputStream pdfcopy = new FileOutputStream(baseOutputFileName+inputFileName); //new File(baseOutputFileName+inputFileName).canRead()
		      
		      OutputStream pdfOutStream = null;
		      
		      String url = "www.caib.es";
		      int position = signer.PDF_SIGN_POSITION_TOP;
		      pdfOutStream = signer.signPDF(pdfin, certificateName, password, contentType, url, position);
		      pdfcopy.write(((ByteArrayOutputStream)pdfOutStream).toByteArray());
		      System.err.println("PDF firmado");
		      		      
		      pdfin.close();
		      pdfOutStream.close();
		      pdfcopy.close();
		      
		      // Compulsa de PDF
		      InputStream pdfin2 = new FileInputStream(inputPath+"/"+inputFileName);
		      OutputStream pdfcopy2 = new FileOutputStream(baseOutputFileName+"_compulsa_"+inputFileName); //new File(baseOutputFileName+inputFileName).canRead()
		      signer.certifyDigitalCopy(pdfin2, pdfcopy2, certificateName, password, contentType, url, "Palma de Mallorca", 0, 0, 0);
		      System.err.println("PDF compulsado");
		      
		      pdfin2.close();
		      pdfcopy2.close();
		      
		} catch(Exception e) {
			
			errores.add(e);
			e.printStackTrace();
			
		}
		
	}


	
	private void verifySmime(String smimeFileName) throws FileNotFoundException, NoSuchProviderException, NoSuchAlgorithmException, CertStoreException, InvalidAlgorithmParameterException, InstantiationException, IllegalAccessException, ClassNotFoundException, MessagingException, IOException, CMSException, NoSuchStoreException, SignatureException {
	      //verificamos el smime
		SMIMEParser smime=signer.getSMIMEParser(new FileInputStream(smimeFileName));
		Signature sigVec[]=smime.getSignatures();
		for(int i=0;i<sigVec.length;i++){
			Signature signature=sigVec[i];
			try {
				if(signature.verify()){
					System.out.println("El certificado de "+signature.getCertSubjectCommonName()+" es válido");
				}else{
					System.out.println("El certificado de "+signature.getCertSubjectCommonName()+" no es válido");
				}
			} catch (SignatureVerifyException e) {
				System.out.println("El certificado de "+signature.getCertSubjectCommonName()+" no se ha podido validar");
				e.printStackTrace();
			}
			
			
			try {
				if(signature.verify(new ByteArrayInputStream((byte[])smime.getSignedObject()))){
					System.out.println("La firma de "+signature.getCertSubjectCommonName()+" es válida");					
				}else{
					System.out.println("La firma de "+signature.getCertSubjectCommonName()+" no es válida");
				}
			} catch (Exception e) {
				System.out.println("La firma de "+signature.getCertSubjectCommonName()+" no se ha podido validar");
			}
			
		}		
	}



	private void runRAW(String certificateName,String password,String additionalAttrib) {
		 //Ejecutamos el circuito para un certificado en concreto
		System.err.println("********* RUN RAW ****************************");
		System.err.println("content-type: "+contentType);	
		System.err.println("certificado para firmar: "+certificateName);
		System.err.println("atributos: "+additionalAttrib);
		
		try {
			check_certs();
		
			 //Firmamos el documento de prueba
			 String baseOutputFileName=outputPath+"/"+version+"_"+
			  		System.getProperty("os.name")+"_"+
			  		System.getProperty("java.specification.version")+"_"+additionalAttrib+"_"+certificateName.replaceAll(":", "").replaceAll("\\\\", "_")+"_";
			  
			  String inputFileName="documento_de_prueba.txt";
			  
			  //generamos el raw
			  InputStream inStream=new MIMEInputStream(new FileInputStream(inputPath+"/"+inputFileName),contentType);
			  
			  //firmamos. el content type indica que viene en raw (base 64)
			  Signature signatureData=signer.sign(inStream, certificateName, password, contentType);
			  sigVec.add(signatureData);
			  inStream.close();
			  
			  //Generamos el fichero de firma
		      ObjectOutputStream oout = new ObjectOutputStream(new FileOutputStream(baseOutputFileName+inputFileName+".firma"));
		      oout.writeObject(signatureData);
		      oout.close();
		      System.err.println("Firma generada");
		      
		      
		      //Validamos el documento raw con el fichero de firma 
		      	//generamos el raw
			  	inStream=new MIMEInputStream(new FileInputStream(inputPath+"/"+inputFileName),contentType);
			  
		      ObjectInputStream oin=new ObjectInputStream(new FileInputStream(baseOutputFileName+inputFileName+".firma"));
		      signatureData = (Signature) oin.readObject();
		      if(signer.verify(inStream, signatureData)){
		    	  System.err.println("Firma verificada"); 
		      }else{
		    	  System.err.println("Error en la validacion");
		      }
		      oin.close();
		      inStream.close();
		      
		      //Generamos el SMIME correspondiente al documento
		      InputStream smimein=new MIMEInputStream(new FileInputStream(inputPath+"/"+inputFileName),contentType);
		      OutputStream smimeout=new FileOutputStream(baseOutputFileName+inputFileName+".p7m.slc");
		      signer.generateSMIME(smimein, signatureData, smimeout);
		      smimein.close();
		      smimeout.close();
		      
		      System.err.println("SMIME generado\n");


		} catch (CircuitException e) {
			errores.add(e);
		} catch (Exception e) {
			errores.add(e);
		}
	}

	

	public CircuitoDePrueba(String [] args){
		parseParams(args);

	    

	}
	
	public void check_certs() throws CircuitException{
		//Inicializamos la api de firma 
	   	//Obtenemos la lista de certificados
	    initSignature();
	    	
	    int i;
	    System.err.print("Certificados instalados: ");
	    for(i=0;i<systemCertList.length;i++){
	    	
	    	System.err.print(" "+systemCertList[i]+((i==systemCertList.length-1)?"":","));
	    }
	    System.err.println("");
	    //Por cada certificado que ha introducido el usuario
	    //miramos si tiene el certificado correspondiente en la lista de certificados
	    Iterator it=circuitCertList.keySet().iterator();
	    while(it.hasNext()){
    		String circuitItem=(String)it.next();
	    	
	    	boolean found=false;
	    	
	    	for(i=0;i<systemCertList.length;i++){
	    		
	    		if(systemCertList[i].equals(circuitItem)){
		    		found=true;
	    		}
	    	}
	    	if(found){
	    		//System.err.println("FOUND: "+circuitItem);
    		}else{
	    		//System.err.println("NOT FOUND: "+circuitItem);
	    		//it.remove();
    		}
    	}

	    
	}
	
	

	
	private  void parseParams(String[] args) {
		int i;
		for(i=0;i<args.length;i++){
			if(args[i].indexOf("-version:")!=-1){
				version=args[i].split(":")[1];
			}else if(args[i].indexOf("@")!=-1){
				String [] cert=args[i].split("@");
				circuitCertList.put(cert[0],(cert.length>1)?cert[1]:"");
			}else if(args[i].indexOf("-in:")!=-1){
				inputPath=args[i].split(":")[1];
			}else if(args[i].indexOf("-out:")!=-1){
				outputPath=args[i].split(":")[1];
			}else if(args[i].indexOf("-fases:")!=-1){
				fases=Integer.parseInt(args[i].split(":")[1]);
			}else{
			
			}
		}
		
	}
	
	
	
	private void initSignature() throws CircuitException
	{
		
		if(signer==null)
			signer = ClientSignerFactory.getSigner();
		
		if (signer == null)
		{
			throw new CircuitException(100,new Throwable("no signer"));
		}
		    
		try {
			systemCertList = signer.getCertList(contentType);
	    }
	    
		catch (SignaturePrivKeyException ex)
	    {
			ex.printStackTrace();
			throw new CircuitException(110,ex);
		}
		catch (SignatureCertNotFoundException cex)
		{
			cex.printStackTrace();
			throw new CircuitException(120,cex);
		}
	
		if (systemCertList==null){
			throw new CircuitException(130,new Throwable("certListNull"));
		}
	}

}
